//
//  SSHTMLFilter.java
//  SwitchableString52
//
//  Created by Paul Suh on Mon Sep 08 2003.
//
/*
 
 Copyright (c) 2003 Paul Suh. All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 
 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.
 
 * Neither the name of the site goodeast.com nor the names of its
 contributors may be used to endorse or promote products derived from
 this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 */

package com.goodeast.webobjects.switchablestring;

import com.webobjects.foundation.*;
import java.util.*;
import java.text.*;


/** 
 * Implements the base functionality for a scheme to filter
 * out HTML tags based on a test. Subclasses need to implement the method
 * tagShouldBeEscaped() which takes a String containing the tag and
 * returns true or false depending on whether or not the tag should be
 * escaped by using &amp;amp; HTML entity or not.
 */
public abstract class SSHTMLFilter {
    
    /**
        Factory class that implements a caching scheme to avoid the overhead
        of creating a new object each time a new filter is desired
        with the same allowed or disallowed list.  
    */
    public static class Factory extends Object {
    
        private static NSMutableDictionary _allowedTagsObjectCache = new NSMutableDictionary();
        private static NSMutableDictionary _disallowedTagsObjectCache = new NSMutableDictionary();
        private static SSHTMLFilter _noTagsObject;
    
        // here to keep it out of the javadoc, and also to keep people
        // from trying to instantiate one of these objects. 
        private Factory() {
            super();
        }
        
	/** 
	 * The method that you should use if you want an SSHTMLFilter object.
	 * It parses the tag strings and returns the correct filter for
	 * the tag strings from the cache if it is already there, or
	 * creates a new subclass of the right type if it is not.
	 * 
	 * @param aTagsString List of allowed tags
	 * @param dTagsString List of disallowed tags
	 * @return An SSHTMLFilter object
	 */
        synchronized public static SSHTMLFilter filterForStrings( String aTagsString, String dTagsString ) {
    
            SSHTMLFilter temp;
            // check the cache first
            if ( isEmptyString( aTagsString ) && isEmptyString( dTagsString ) ) {
                if ( _noTagsObject == null )
                    _noTagsObject = new SSNoTagsAllowedFilter();
                return _noTagsObject;
            } else if ( ( temp = (SSHTMLFilter) _allowedTagsObjectCache.objectForKey( aTagsString ) ) != null )
                return temp;
            else if ( ( temp = (SSHTMLFilter) _disallowedTagsObjectCache.objectForKey( dTagsString ) ) != null )
                return temp;

            NSArray aTags = parseTagString( aTagsString );
            NSArray dTags = parseTagString( dTagsString );
                
            if ( aTags.count() == 0 && dTags.count() == 0 ) {
                
                if ( _noTagsObject == null )
                    _noTagsObject = new SSNoTagsAllowedFilter();
                return _noTagsObject;
                
            } else if ( aTags.count() > 0 ) {
                
                if ( ( temp = (SSHTMLFilter) _allowedTagsObjectCache.objectForKey( aTagsString ) ) == null ) {
                    temp = new SSAllowedTagsFilter( parseTagString( aTagsString ) );
                    _allowedTagsObjectCache.setObjectForKey( temp, aTagsString );
                }
                return temp;
                
            } else if ( dTags.count() > 0 ) {
    
                if ( ( temp = (SSHTMLFilter) _disallowedTagsObjectCache.objectForKey( dTagsString ) ) == null ) {
                    temp = new SSAllowedTagsFilter( parseTagString( dTagsString ) );
                    _disallowedTagsObjectCache.setObjectForKey( temp, dTagsString );
                }
                return temp;
    
            } else {
    
                if ( NSLog.debugLoggingAllowedForLevel( NSLog.DebugLevelCritical ) ) {
                    NSLog.err.appendln( "SSHTMLFilterFactory: bad set of filter strings received." );
                    NSLog.err.appendln( "SSHTMLFilterFactory: aTagString = " + aTagsString );
                    NSLog.err.appendln( "SSHTMLFilterFactory: dTagString = " + dTagsString );
                    (new RuntimeException()).printStackTrace();
                }
                
                if ( _noTagsObject == null )
                    _noTagsObject = new SSNoTagsAllowedFilter();
                return _noTagsObject;
            }
        }
        
	/** 
	 * Utility method to determine if a string is empty or null.
	 * 
	 * @param s String to be tested
	 * @return True if the string is empty or null
	 */
        private static boolean isEmptyString( String s ) {
            
            if ( s == null ) return true;
            if ( s.length() == 0 ) return true;
            return false;
            
        }
        
        
	/** 
	 * Utility method that parses a list of tags into an NSArray.
	 * Both simple comma-separated lists or plist-style lists are allowed.
	 * Indicators are: <UL>
	 * <LI>Plist: first character is "("</LI>
	 * <LI>XML-style plist: first 5 characters are "&lt;?xml"</LI>
	 * <LI>Simple list: anything else</LI>
	 * </UL>
	 * 
	 * @param tagString String containing list of HTML tags
	 * @return List of HTML tags
	 */
    
        private static NSArray parseTagString( String tagString ) {
    
            try {
                if ( ( tagString == null ) || ( tagString.equals( "" ) ) )
                    return new NSArray();
                else if ( tagString.startsWith( "<?xml" ) || tagString.startsWith( "(" ) )
                    return NSPropertyListSerialization.arrayForString( tagString );
                else {
                    // strip out all of the spaces
                    // strip out "<" and ">"
                    // uppercase the whole thing
                    // split on ","
                    String temp = NSArray.componentsSeparatedByString( tagString, " " ).componentsJoinedByString( "" );
                    temp = NSArray.componentsSeparatedByString( temp, "<" ).componentsJoinedByString( "" );
                    temp = NSArray.componentsSeparatedByString( temp, ">" ).componentsJoinedByString( "" );
                    return NSArray.componentsSeparatedByString( temp.toUpperCase(), "," );
                }
            } catch ( Throwable t ) {
    
                if ( NSLog.debugLoggingAllowedForLevel( NSLog.DebugLevelCritical ) ) {
                    NSLog.err.appendln( "SSHTMLFilterFactory: unparseable filter string received." );
                    NSLog.err.appendln( "SSHTMLFilterFactory: filter string = " + tagString );
                    t.printStackTrace();
                }
                return new NSArray();
                
            }
        }
        
    }
    
    /**
     * Used by SSHTMLFilter subclasses. 
     */
    protected SSHTMLFilter() {
        super();
    }

    /** 
     * This is a complex method that parses the raw string looking for
     * tags. Once a tag is found, it is passed to method tagShouldBeEscaped()
     * to determine whether or not it will be escaped. It is a LALR(2) parser
     * at its core, based on a 3 item FIFO buffer along with a stack that
     * keeps track of which tags are currently open so that they can be
     * closed at the end of the string if needed. The state machine has two
     * states: outside a tag vs. inside a tag. It handles HTML comments and
     * entities (stuff that starts with &amp;!) as well as regular tags. The
     * raw string is split into pieces based on the characters "&lt;", "/",
     * "!", and "&gt;". <BR>
     * <BR>
     * The algorithm in detail is:<BR>
     * 
     * three token FIFO buffer A1, A2, A3<BR>
     * stack B<BR>
     * output buffer result<BR>
     * 
     * tokenize the raw string splitting on "&lt;", "/", "!", and "&gt;". Include
     * the split characters as tokens.
     * for each token
     * <BLOCKQUOTE>
     * append A3 to result<BR>
     * shift A2 to A3<BR>
     * shift A1 to A2<BR>
     * load next token into A1<BR>
     * <BR>
     * if inside a tag
     * <BLOCKQUOTE>
     * if A1 == ">"
     * <BLOCKQUOTE>
     * change state to outside a tag
     * </BLOCKQUOTE>
     * </BLOCKQUOTE>
     * else if outside a tag
     * <BLOCKQUOTE>
     * if A2 == "<" and A1 !="/"
     * <BLOCKQUOTE>
     * if A1.firstWord in disallowedList or A1.firstWord not in allowedList
     * <BLOCKQUOTE>
     * change A2 to "&lt;"
     * </BLOCKQUOTE>
     * else if A1.firstWord is not BR or LI
     * <BLOCKQUOTE>
     * push A1.firstWord onto B
     * </BLOCKQUOTE>
     * change state to inside a tag
     * </BLOCKQUOTE>
     * 
     * if A3 == "<" and A2 == "/"
     * <BLOCKQUOTE>
     * if A1.firstWord in disallowedList or A1.firstWord not in allowedList
     * <BLOCKQUOTE>
     * change A3 to "&lt;"
     * </BLOCKQUOTE>
     * else if B.top == A2.firstWord
     * <BLOCKQUOTE>
     * pop B
     * </BLOCKQUOTE>
     * change state to inside a tag
     * </BLOCKQUOTE>
     * </BLOCKQUOTE>
     * </BLOCKQUOTE>
     * After all tokens have been processed, we need to deal with edge cases
     * after the end of the loop. <BR>
     * <BR>
     * if outside a tag
     * <BLCOKQUOTE>
     * if A1 == "<"
     * <BLOCKQUOTE>
     * convert A1 to "&lt;"
     * </BLOCKQUOTE>
     * if A2 == "<" and A1 == "/"
     * <BLOCKQUOTE>
     * convert A2 to "&lt;"
     * </BLOCKQUOTE>
     * if A2 == "<" and A1 == "!"
     * <BLOCKQUOTE>
     * convert A2 to "&lt;"
     * </BLOCKQUOTE>
     * </BLOCKQUOTE>
     * flush A1, A2 A3 to result<BR>
     * <BR>
     * if inside a tag
     * <BLOCKQUOTE>
     * append ">" to result
     * </BLOCKQUOTE>
     * pop elements off of B and create close tags for them<BR>
     * <BR>
     * Note: some tags (like &lt;BR>) don't have a closing tag, but these get closed off
     * anyway, since it doesn't hurt in general and it makes the algorithm much
     * simpler. BR and LI are special cased, since they are in the Slashdot set
     * of allowed tags and we're going to recommend that set to developers.
     * 
     * @param rawString String containing possible undesired and/or malformed 
     *     HTML
     * @return String with undesired HTML tags escaped out
     */

    public String filterForAllowedHTML( String rawString ) {

        // split the string using a StringTokenizer on "<" or "/", include both these as
        // tokens
        StringTokenizer st = new StringTokenizer( rawString, "</!>", true );

        // pre-allocating a buffer may be an unnecessary optimization, but it can't hurt
        StringBuffer result = new StringBuffer( rawString.length() + 256 );

        String A1 = null;	// FIFO queue elements
        String A2 = null;
        String A3 = null;

        NSMutableArray B = new NSMutableArray();	// stack

        boolean tagOpen = false;

        while ( st.hasMoreTokens() ) {

            if ( A3 != null )
                result.append( A3 );
            A3 = A2;
            A2 = A1;
            A1 = st.nextToken();

            if ( tagOpen ) {

                if ( A1.equals( ">" ) )
                    tagOpen = false;

            } else {


                if ( ( A2 != null ) && ( A2.equals("<") ) && !A1.equals( "/" ) ) {
                    // if A2 != null then A1 cannot be null

                    if ( !Character.isLetter( A1.charAt( 0 ) ) ) {
                        // if first character is not a letter, then it can't be part of a tag
                        // so knock it out
                        // this handles the case of a <!--  --> comment

                        A2 = "&lt;";

                    } else {

                        BreakIterator bi = BreakIterator.getWordInstance();
                        bi.setText( A1 );
                        String firstWord = A1.substring( 0, bi.next() ).toUpperCase();

                        if ( this.tagShouldBeEscaped( firstWord ) )
                            A2 = "&lt;";
                        else {
                            // special case BR and LI tags since they don't need
                            // closing tags
                            if ( !( firstWord.equals( "BR" ) || firstWord.equals( "LI" ) ) )
                                B.addObject( firstWord );
                            tagOpen = true;
                        }

                    }

                } else if ( ( A3 != null ) && ( A3.equals( "<" ) ) && ( A2.equals( "/" ) ) ) {
                    // if A3 != null then A1 and A2 cannot be null

                    // if first character is not a letter, then it can't be part of a tag
                    if ( Character.isLetter( A1.charAt( 0 ) ) ) {

                        BreakIterator bi = BreakIterator.getWordInstance();
                        bi.setText( A1 );
                        String firstWord = A1.substring( 0, bi.next() ).toUpperCase();

                        if ( this.tagShouldBeEscaped( firstWord ) )
                            A3 = "&lt;";
                        else {
                            if ( B.lastObject().equals( firstWord ) )
                                B.removeLastObject();
                            tagOpen = true;
                        }
                    }

                }
            }
        }


        // end of loop, handle edge cases
        if ( !tagOpen && ( A1 != null ) && A1.equals( "<" ) )
            A1 = "&lt;";
        else if ( !tagOpen && ( A2 != null ) && A2.equals( "<" ) ) {
            if ( A1.equals( "/" ) )
                A2 = "&lt;";
            else if ( A1.equals( "!" ) )
                A2 = "&lt;";
        }

        // append remainder
        if ( A3 != null )
            result.append( A3 );
        if ( A2 != null )
            result.append( A2 );
        if ( A1 != null )
            result.append( A1 );

        // close any open tags
        if ( tagOpen )
            result.append( ">" );

        // close any dangling tags
        Enumeration e = B.reverseObjectEnumerator();
        while ( e.hasMoreElements() ) {

            result.append( "</" + e.nextElement() + ">" );

        }
        return result.toString();
    }

    /** 
     * Method called to determine if a particular tag should be escaped or not. Subclasses
     * must override this method to provide the correct behavior.
     * 
     * @param tag Tag to be checked
     * @return True is tag is undesired, false if it is OK
     */
    public abstract boolean tagShouldBeEscaped( String tag );
    
}
